import functools
import requests
import json

s = requests.Session()


@functools.lru_cache(maxsize=16)
def get_json(url, **kwargs):
    """
    Auxiliary function to make get requests
    within a HTTP Session.

    Parameters
    ----------
    url : str
        Url to be requested.

    **kwargs
        Passed to Session().get()

    Returns
    -------
    dict
        Decoded JSON.
    """
    response = s.get(url, timeout=60, **kwargs)
    response.raise_for_status()
    try:
        return response.json()
    except json.JSONDecodeError:
        raise ValueError(f"A request to {url} didn't produce any JSON.")



from pandas import DataFrame
from .request import get_json

metadata_list = ["codigo_sgs", "title", "periodicidade", "unidade_medida"]


def bcb_get_search_results(url):
    """
    Request a BCB search result (CKAN API) and parse it into
    a DataFrame. Get only the columns in `metadata_list`.
    """
    json = get_json(url)
    search_results = json["result"]["results"]
    assert search_results, "Nothing was found."
    return DataFrame(search_results).loc[:, metadata_list]


def ipea_get_search_results(url):
    """
    Request a IPEA search result and parse it into a
    DataFrame.
    """
    json = get_json(url)
    search_results = json["value"]
    assert search_results, "Nothing was found."
    return DataFrame(search_results)



import datetime

year_format = ["%Y"]
month_year_format = ["%m/%y", "%m/%Y", "%M/%Y", "%b/%Y", "%b/%y", "%B/%Y", "%B/%y"]
day_month_year_format = ["%d/%m/%Y"]

slash_formats = year_format + month_year_format + day_month_year_format
dash_formats = [fmt.replace("/", "-") for fmt in slash_formats]
blank_formats = [fmt.replace("/", "") for fmt in slash_formats]

allowed_fmts = slash_formats + dash_formats + blank_formats


def parse_date(date_str, api, start=True):
    """
    Auxiliary function to convert different dates strings
    to the format required by an API.

    Also, when it is an end date, if day is unspecificed,
    the day will be the last day of the month.
    Similarly, if month is not specified, the month must be 12.

    Note that the date string is assumed to start with day
    or month. Otherwise, unexpected results can arise.

    Parameters
    ----------
    date_str : str
        String to be parsed.

    api : str
        Name of the api. Possible values are "ibge", "ipea" and "bcb".

    start : bool
        If it is a start date.

    Returns
    -------
    str
        Appropriate date string for a given API.

    Raises
    ------
    ValueError
        If the string could not be converted to a
        datetime.datetime object by strptime.

    Examples
    --------
    >>> dates.parse_date("012017", api="bcb")
    '01/01/2017'
    >>> dates.parse_date("01-2017", api="ipea")
    '2017-01-01T00:00:00-00:00'
    >>> dates.parse_date("01/2017", api="ibge")
    '201701'
    """
    assert isinstance(date_str, str), "You didn't give a string as a date."
    for fmt in allowed_fmts:
        try:
            date = datetime.datetime.strptime(date_str, fmt)
            if not start:
                # if an end date,
                # "2011"    -> "31-12-2011"
                # "01-2011" -> "31-01-2011"
                # "02-01-2011" -> "02-01-2011"
                date = date.replace(
                    month=12 if fmt in year_format else date.month,
                    day=last_day_of_month(date) if fmt.find("%d") == -1 else date.day,
                )
            return api_date_format_of(date, api)
        except ValueError:
            continue
    raise ValueError("Not a valid date format.")


def parse_dates(start, end, api):
    """
    Auxiliary function to call dates functions.

    Returns
    -------
    str
        Date format as required by an API.
    """
    if start:
        start = parse_date(start, api, start=True)
    else:
        start = api_date_format_of(very_old_date(), api)
    if end:
        end = parse_date(end, api, start=False)
    else:
        end = api_date_format_of(today_date(), api)
    return start, end


def api_date_format_of(date, api):
    """
    Auxiliary function to convert datetime.datetime
    object to string compatible with a given API.
    """
    if api == "ipea":
        return date.strftime("%Y-%m-%dT00:00:00") + "-00:00"
    if api == "bcb":
        return date.strftime("%d/%m/%Y")
    if api == "ibge":
        return date.strftime("%Y%m")


def last_day_of_month(date):
    """
    Auxiliary function that returns the last day of the month,
    if no day is specified.

    See: https://stackoverflow.com/questions/42950/get-last-day-of-the-month
    """
    next_month = date.replace(day=28) + datetime.timedelta(days=4)
    return (next_month - datetime.timedelta(days=next_month.day)).day


def very_old_date():
    """
    Returns very old date (01/01/1900).

    Returns
    -------
    datetime.datetime
    """
    return datetime.datetime(1900, 1, 1)


def today_date():
    """
    Returns today's date.

    Returns
    -------
    datetime.datetime
        Today's date.
    """
    return datetime.datetime.today()


def month_to_quarter(date, fmt=None):
    """
    Convert month to quarter, i.e.,
    12 -> 4, 6 -> 2, 7 -> 3 etc.

    Parameters
    ----------
    date : datetime.datetime

    Returns
    -------
    datetime.datetime

    Examples
    --------
        >>> dates.month_to_quarter(datetime.datetime(2019, 12, 1))
        datetime.datetime(2019, 4, 1, 0, 0)
    """
    if isinstance(date, str) and fmt:
        date = datetime.datetime.strptime(date, fmt)
    quarter = divmod(date.month, 3)[0] + bool(divmod(date.month, 3)[1])
    return date.replace(month=quarter)


def check_if_quarter(dates):
    """
    Check if month of a datetime
    object is less than 4, i.e., if
    it is a quarter.
    """
    for date in dates:
        date_obj = datetime.datetime.strptime(date, "%Y%m")
        if date_obj.month > 4:
            error_msg = f"This is a quarterly time series."
            error_msg += (
                f" {date_obj.month} is not a quarter, choose a number between 1 and 4."
            )
            raise ValueError(error_msg)



import re
from .utils import cat, is_iterable
from .ipea_metadata_list import ipea_metadata_list
from .dates import month_to_quarter, check_if_quarter, parse_dates

# IPEA


def ipea_make_date_query(start=None, end=None):
    """
    Auxiliary function to return the right
    string for filtering dates via IPEA api.

    Parameters
    ----------
    start : str
        Start date string.

    End : str
        End date string.

    Returns
    -------
    str
        A string to filter by dates.

    Examples
    --------
    >>> url.ipea_make_dates_query("2019-01-01T00:00:00-00:00", "2019-02-01T00:00:00-00:00")
    '&$filter=VALDATA ge 2019-01-01T00:00:00-00:00 and VALDATA le 2019-02-01T00:00:00-00:00'
    """
    dates = ""
    if start and end:
        dates = f"&$filter=VALDATA ge {start} and VALDATA le {end}"
    elif start:
        dates = f"&$filter=VALDATA ge {start}"
    elif end:
        dates = f"&$filter=VALDATA le {end}"
    return dates


def ipea_make_select_query(metadatas):
    """
    Auxiliary function to make select
    query for IBGE's web API.

    It loops through the keys of a dictionary
    and makes the string by joining them.

    Parameters
    ----------
    metadatas : dict
        Metadatas used in the search.

    Returns
    -------
    str
        A string to select metadatas.

    Examples
    --------
        >>> url.ipea_make_select_query("")
        '?$select=SERCODIGO,SERNOME,PERNOME,UNINOME'

        >>> url.ipea_make_select_query(["FNTNOME"])
        '?$select=SERCODIGO,SERNOME,PERNOME,UNINOME,FNTNOME'

        If a metadata is already selected by default, nothing changes:

        >>> url.ipea_make_select_query(["PERNOME"])
        '?$select=SERCODIGO,SERNOME,PERNOME,UNINOME'
    """
    # to get the string
    # SERCODIGO,PERNOME,UNINOME,SERNOME,ANOTHERFILTER,ANOTHERFILTER
    # where ANOTHER must be something not alreay selected by default
    defaults = ["SERCODIGO", "SERNOME", "PERNOME", "UNINOME"]
    selected = defaults + [
        metadata for metadata in metadatas if metadata not in defaults
    ]
    return f"?$select={','.join(selected)}"


def ipea_make_filter_query(names, metadatas={}):
    """
    Auxiliary function to make filter
    query for IBGE's database API.

    Parameters
    ----------
    names : list of str
        Strings to filter by name.

    metadatas : dict
        Dictionary whose keys are metadatas
        and values strings (or list of strings)
        to look up for.

    Returns
    -------
    str
        A string to filter metadatas.

    Raises
    ------
    ValueError
        If not a valid metadata.

    Examples
    --------
        >>> url.ipea_make_filter_query("SELIC")
        "&$filter=contains(SERNOME,'SELIC')"

        >>> url.ipea_make_filter_query("SELIC", {"PERNOME": ["mensal", "trimestral"], "FNTNOME": "IBGE"})
        "&$filter=contains(SERNOME,'SELIC') and (contains(PERNOME,'mensal') or contains(PERNOME,'trimestral')) and contains(FNTNOME,'IBGE')"

        >>> url.ipea_make_filter_query("SELIC", {"SERSTATUS": "A", "SERNUMERICA": 1})
        "&$filter=contains(SERNOME,'SELIC') and SERSTATUS eq 'A' and SERNUMERICA eq 1"
    """
    # friendly error message in case of invalid metadata
    invalid_metadatas = [
        metadata for metadata in metadatas if metadata not in ipea_metadata_list
    ]
    error_msg = f"{', '.join(invalid_metadatas)}: non-valid metadata. Call ipea.list_metadata() if you need help."
    if invalid_metadatas:
        raise ValueError(error_msg)

    # building filter query
    prefix = "&$filter="

    # string to filter by name
    filter_name = contains_operator("SERNOME", names, " and ") if names else ""

    # start building string to filter by additional metadata
    filter_metadata = ""
    if metadatas:
        metadata_filters = []
        for metadata, value in metadatas.items():
            # if metadata is numeric, use OData equal operator
            if re.search("(CODIGO|NUMERICA|STATUS)$", metadata):
                metadata_filters.append(equal_operator(metadata, value))
            # else, use contains operator
            else:
                metadata_filters.append(contains_operator(metadata, value))
        # if user filtered by name, prepend metadatas filter with an 'and'
        filter_metadata = " and " if filter_name else ""
        filter_metadata += " and ".join(metadata_filters)
    return f"{prefix}{filter_name}{filter_metadata}"


def contains_operator(metadata, values, logical_operator=" or "):
    """
    Auxiliary function to make string with
    OData's contains logical operator.

    Parameters
    ----------
    metadata : str or list of str
        Metadatas to be filtered.

    metadata : str or list of str
        Values to filter by.

    Returns
    -------
        A valid string to perform the query
        via OData's URL convention.

    Examples
    --------
    >>> url.contains_operator("FNTNOME", ["A", "B"])
    "(contains('FNTNOME','A') or contains('FNTNOME','B'))"
    """
    if isinstance(values, (list, tuple)):
        filters = [f"contains({metadata},'{item}')" for item in values]
        filters_joined = logical_operator.join(filters)
        return f"({filters_joined})"
    else:
        return f"contains({metadata},'{values}')"


def equal_operator(metadata, values):
    """
    Auxiliary function to make string with
    OData's equal logical operator.

    Parameters
    ----------
    metadata : str or list of str
        Metadatas to be filtered.

    metadata : str or list of str
        Values to filter by.

    Returns
    -------
        A valid string to perform the query
        via OData's URL convention.

    Examples
    --------
    >>> url.equal_operator("SERNUMERICA", [1, "A"])
    "('SERNUMERICA'' eq 1 or 'SERNUMERICA', eq 'A')"
    """
    if isinstance(values, (list, tuple)):
        filters = [f"{metadata} eq {quote_if_str(item)}" for item in values]
        return f"({' or '.join(filters)})"
    else:
        return f"{metadata} eq {quote_if_str(values)}"


def quote_if_str(something):
    """
    Auxiliary function to put quotes around value
    if it is a string. Needed to make filter queries
    for OData's equal logical operator in case of a
    string.

    Parameters
    ----------
    something
        Any object.

    Returns
    -------
        A string around quote if something is a string,
        else just the object coerced to a string.
    """
    return f"'{something}'" if isinstance(something, str) else f"{something}"


# IBGE


def ibge_make_classifications_query(classifications=None):
    """
    Auxiliary function to make classifications
    part of the URL.

    Parameters
    ----------
    classifications : int, str, list or dict
        Dictionary of classifications (keys) and categories
        (values) or a set of classifcations as int, str, or
        list.

    Returns
    -------
    str
        A valid string to filter by classifications
        and categories.

    Examples
    --------
    >>> url.ibge_make_classification_query({1: [2, 3]})
    'classificacao=1[2,3]'

    >>> url.ibge_make_classification_query([1, 2])
    'classificacao=1[all]|2[all]'

    >>> url.ibge_make_classification_query(3)
    'classificacao=3[all]'
    """
    if isinstance(classifications, dict):
        s = []
        for classification, category in classifications.items():
            if not category or category == "all":
                s.append(f"{classification}[all]")
            else:
                s.append(f"{classification}[{cat(category, ',')}]")
        return "classificacao=" + "|".join(s)
    elif isinstance(classifications, (int, str)):
        return f"classificacao={classifications}[all]"
    elif isinstance(classifications, list):
        piped_classifications = "|".join(
            [f"{classification}[all]" for classification in classifications]
        )
        return f"classificacao={piped_classifications}"
    else:
        return ""


def ibge_make_dates_query(start=None, end=None, last_n=None, freq=None):
    """
    Auxiliary function to filter a time series'
    periods.

    Parameters
    ----------
    start : str
        Initial date string.

    end : str
        Final date string.

    last_n : str or int
        Get last n observations

    freq : str
        Time series frequency.

    Returns
    -------
    str
        A valid string to filter dates in IBGE's API.

    Examples
    --------
    >>> url.ibge_make_dates_query(last_n=5)
    '/periodos/-5'
    >>> url.ibge_make_dates_query(start="012017")
    '/periodos/201701-201912'
    >>> url.ibge_make_dates_query(end="072017")
    '/periodos/190001-201707'
    >>> url.ibge_make_dates_query(start="052015", end="072017")
    '/periodos/201505-201707'
    """
    start, end = parse_dates(start, end, "ibge")
    if last_n:
        return f"/periodos/-{last_n}"
    if freq == "trimestral":
        start = month_to_quarter(start, "%Y%m").strftime("%Y%m")
        end = month_to_quarter(end, "%Y%m").strftime("%Y%m")
        check_if_quarter((start, end))
    elif freq == "anual":
        start, end = start[:-2], end[:-2]
    return f"/periodos/{start}-{end}"


def ibge_make_variables_query(variables=None):
    """
    Auxiliary function to filter an IBGE's
    aggregate by variables.

    Parameters
    ----------
    variables : int or list of int
        The variables' codes.

    Returns
    -------
    str
        A string to filter variables in IBGE's API.

    Examples
    --------
    >>> url.ibge_make_variables_query(100)
    '/variaveis/100'
    >>> url.ibge_make_variables_query([1, 2, 3])
    '/variaveis/1|2|3'
    >>> url.ibge_make_variables_query()
    '/variaveis/'
    """
    if is_iterable(variables):
        return f"/variaveis/{cat(variables, '|')}"
    elif variables:
        return f"/variaveis/{variables}"
    else:
        return f"/variaveis"


locations_dict = {
    "N6": "municipality",
    "N3": "state",
    "N2": "macroregion",
    "N7": "mesoregion",
    "N9": "microregion",
    "N1": "brazil",
}


location_ids = {location: code for code, location in locations_dict.items()}


def ibge_make_locations_query(
    municipality=None,
    state=None,
    macroregion=None,
    microregion=None,
    mesoregion=None,
    brazil=None,
):
    """
    Auxiliary function to filter an IBGE's
    aggregate by variables.

    Parameters
    ----------
    municipality
    state
    macroregion
    microregion
    mesoregion
    brazil

    Returns
    -------
    str
        A string to filter locations in IBGE's API.

    Examples
    --------
    >>> url.ibge_make_location_query()
    '&localidades=BR'
    >>> url.ibge_make_location_query(city=True)
    '&localidades=N6'
    >>> url.ibge_make_location_query(city=1)
    '&localidades=N6[1]'
    >>> url.ibge_make_location_query(city=[2, 3, 4])
    '&localidades=N6[2,3,4]'
    """
    # note-to-self: http://api.sidra.ibge.gov.br/desctabapi.aspx?c=136
    locations = vars()
    prefix = "&localidades="
    query = []
    if all([code is None for code in locations.values()]):
        return prefix + "BR"
    for location, codes in locations.items():
        if location == "brazil" and codes:
            query.append("BR")
        elif isinstance(codes, list):
            query.append(f"{location_ids[location]}[{cat(codes, ',')}]")
        elif type(codes) == int:
            query.append(f"{location_ids[location]}[{codes}]")
        elif codes:
            query.append(f"{location_ids[location]}")
    return prefix + "|".join(query)


# vi: nowrap



import pandas as pd
from .request import get_json


def bcb_json_to_df(url, code, name):
    """
    Parse a timeseries returned by BCB API
    into a DataFrame.
    """
    json = get_json(url)
    df = pd.DataFrame(json)
    df["valor"] = df["valor"].astype("float64")
    df = df.set_index("data")
    df = df.rename_axis("Date")
    df.index = pd.to_datetime(df.index, format="%d/%m/%Y")
    df.columns = [name if name else code]
    return df


def ipea_json_to_df(url, code, name):
    """
    Parse a timeseries returned by IPEA API
    into a DataFrame.
    """
    json = get_json(url)["value"]
    df = pd.DataFrame(json)

    # removing utc component from date string
    try:
        df["VALDATA"] = df["VALDATA"].str[:-6]
        df = df.set_index("VALDATA")
        df = df.rename_axis("Date")
        df.index = pd.to_datetime(df.index, format="%Y-%m-%dT%H:%M:%S")
        # casting numerical values
        df["VALVALOR"] = pd.to_numeric(df["VALVALOR"], errors="coerce")
        df.columns = [name if name else code]
    except KeyError:
        return

    return df


def ibge_json_to_df(url, freq="mensal"):
    """
    Parse a IBGE table/aggregate in JSON format
    into a DataFrame.
    """
    json = get_json(url)
    # ignore first element as it as no real data
    df = pd.DataFrame(json[1:])
    # getting columns names
    df.columns = json[0].values()

    # handling dates
    date_fmt = "%Y" if freq == "anual" else "%Y%m"
    date_key = json[0]["D2C"]
    df[date_key] = pd.to_datetime(df[date_key], format=date_fmt)
    df = df.set_index(date_key)
    df = df.rename_axis("Date")

    # handling numerical values
    df["Valor"] = pd.to_numeric(df["Valor"], errors="coerce")

    # getting location code name column
    location_key = json[0]["D1C"]

    # dropping less useful columns
    # i.e., any columns which represents code,
    # except for the location and variable code,
    # and a bunch of other columns
    df = df.drop(
        [
            c
            for c in df.columns
            if c.endswith("(Código)") and c not in [location_key, "Variável (Código)"]
        ]
        + ["Mês", "Unidade de Medida", "Brasil", "Nível Territorial"],
        axis="columns",
        errors="ignore",
    )
    return df



ipea_metadata_list = {
    "SERNOME": "Name",
    "SERCODIGO": "Code",
    "PERNOME": "Frequency",
    "UNINOME": "Unit of measurement",
    "BASNOME": "Basis's name",
    "TEMCODIGO": "Theme's code",
    "PAICODIGO": "Country / Region's code",
    "SERCOMENTARIO": "Comments/Notes",
    "FNTNOME": "Source's name",
    "FNTSIGLA": "Source's initials",
    "FNTURL": "Source's url",
    "MULNOME": "Multiplicative factor",
    "SERATUALIZACAO": "When it was last updated",
    "SERSTATUS": "Active ('A'), Inactive ('I')",
    "SERNUMERICA": "Numeric (1), Alphanumeric (0)",
}



import pandas as pd

from .helpers.request import get_json
from .helpers.response import ibge_json_to_df
from .helpers.metadata import ibge_metadata_to_df
from .helpers.utils import search_df
from .helpers.lists import list_regions_helper
from .helpers.url import (
    ibge_make_dates_query,
    ibge_make_variables_query,
    ibge_make_locations_query,
    ibge_make_classifications_query,
    locations_dict,
)


def get_series(
    code,
    variables=None,
    start=None,
    end=None,
    last_n=None,
    municipalities=None,
    states=None,
    macroregions=None,
    microregions=None,
    mesoregions=None,
    brazil=None,
    classifications=None,
):
    """
    Function to get variables associated with an
    aggregate from IBGE's database.

    Parameters
    ----------
    code : int
        Aggregate's code.

    variables : int or list of ints, optional
        Which variables to select (if None, return all of them).

    start : int or str, optional
        Initial date, month or day first.

    end : int or str, optional
        Final date, month or day first.

    last_n : int or str, optional
        Return only last n observations.

    municipalities : str, int, bool a list, optional
        Municipalities' codes.

    states : str, int, bool or a list, optional
        States' codes.

    macroregions : str, int, bool or a list, optional
        Macroregions' codes.

    microregions : str, int, bool or a list, optional
        Microregions' codes.

    mesoregions : str, int, bool or a list, optional
        Mesoregions' codes.

    classifications : dict, int, str or list, optional
        Classifications' / categories' codes.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with series values and metadata.

    Examples
    --------
    >>> ibge.get_series(1419, start="11-2019", end="11-2019")
                Nível Territorial                              Variável   Geral, grupo, subgrupo, item e subitem   Valor
    Date
    2019-11-01            Brasil                 IPCA - Variação mensal   Índice geral                              0.51
    2019-11-01            Brasil       IPCA - Variação acumulada no ano   Índice geral                              3.12
    2019-11-01            Brasil  IPCA - Variação acumulada em 12 meses   Índice geral                              3.27
    2019-11-01            Brasil                     IPCA - Peso mensal   Índice geral                            100.00
    """
    baseurl = f"https://servicodados.ibge.gov.br/api/v3/agregados/{code}"
    frequency = get_frequency(code)
    dates = ibge_make_dates_query(start, end, last_n, frequency)
    variables = ibge_make_variables_query(variables)
    locations = ibge_make_locations_query(
        municipalities, states, macroregions, microregions, mesoregions, brazil
    )
    classifications = ibge_make_classifications_query(classifications)
    url = f"{baseurl}{dates}{variables}?{classifications}{locations}&view=flat"
    return ibge_json_to_df(url, frequency)


def get_frequency(aggregate):
    """
    Auxiliary function to get frequency of
    a time series from IBGE's database.

    This is needed because in case of a yearly time
    series, there can't be months in the url.
    """
    return list_periods(aggregate).loc["frequencia", :].values


def get_metadata(aggregate):
    """
    Get metadata of a time series from IBGE's database.

    Parameters
    ----------
    aggregate : str
        Aggregate's code.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with metadata values.

    Examples
    --------
    >>> ibge.get_metadata(1419)
                                                                 values
    id                                                             1419
    nome              IPCA - Variação mensal, acumulada no ano, acum...
    URL                            http://sidra.ibge.gov.br/tabela/1419
    pesquisa              Índice Nacional de Preços ao Consumidor Amplo
    assunto                                           Índices de preços
    periodicidade     {'frequencia': 'mensal', 'inicio': 201201, 'fi...
    nivelTerritorial  {'Administrativo': ['N1', 'N6', 'N7'], 'Especi...
    variaveis         [{'id': 63, 'nome': 'IPCA - Variação mensal', ...
    classificacoes    [{'id': 315, 'nome': 'Geral, grupo, subgrupo, ...
    """
    url = f"https://servicodados.ibge.gov.br/api/v3/agregados/{aggregate}/metadados"
    return ibge_metadata_to_df(url)


def search(*search, **searches):
    """
    Function to list all aggregates in IBGE's database.

    Parameters
    ----------
    *search
        Strings to search in aggregates' names.

    **searches
        Strings to search in other columns.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with an aggregate's metadata.

    Examples
    --------
    >>> ibge.search("Índice", "Preços", pesquisa_nome="Pesquisa").head()
            id                                               nome pesquisa_id                              pesquisa_nome
    2472  1399  Número de empresas comerciais, Unidades locais...          PB                 Pesquisa Anual de Comércio
    2957   379  Índice de Gini - recebimento médio mensal das ...          OF          Pesquisa de Orçamentos Familiares
    3101    50  Folha de pagamento nominal por classes de indú...          DG  Pesquisa Industrial Mensal - Dados Gerais
    3102    49  Folha de pagamento nominal por tipo de índice ...          DG  Pesquisa Industrial Mensal - Dados Gerais
    3103    52  Folha de pagamento nominal por trabalhador por...          DG  Pesquisa Industrial Mensal - Dados Gerais
    """
    json = get_json("https://servicodados.ibge.gov.br/api/v3/agregados")
    df = pd.io.json.json_normalize(
        json, record_path="agregados", meta=["id", "nome"], meta_prefix="pesquisa_"
    )
    if search or searches:
        return search_df(df, search, searches).reset_index(drop=True)
    return df


def list_variables(aggregate, *search, **searches):
    """
    Function to list all variables associated with an aggregate.

    Parameters
    ----------
    aggregate : int or str
        Aggregate's code.

    *search
        Names to search for.

    **searches
        Strings to search in other columns.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with all available variables of an aggregate.

    Examples
    --------
    >>> ibge.list_variables(1419)
         id                               variavel unidade
    0    63                 IPCA - Variação mensal       %
    1    69       IPCA - Variação acumulada no ano       %
    2  2265  IPCA - Variação acumulada em 12 meses       %
    3    66                     IPCA - Peso mensal       %
    """
    baseurl = "https://servicodados.ibge.gov.br/api/v3"
    query = f"/agregados/{aggregate}/variaveis/all?localidades=BR"
    url = f"{baseurl}{query}"
    json = get_json(url)
    df = pd.io.json.json_normalize(json).iloc[:, :3]
    if search or searches:
        return search_df(df, search, searches).reset_index(drop=True)
    return df


def list_locations(aggregate):
    """
    Function to list locations of a given aggregate.

    Parameters
    ----------
    aggregate : int or str
        Aggregate's code.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the available locations for an aggregate.

    Examples
    --------
    >>> ibge.list_locations(1419)
      codes  parameters
    0    N1      brazil
    1    N6        city
    2    N7  mesoregion
    """
    baseurl = "https://servicodados.ibge.gov.br/api/v3"
    query = f"/agregados/{aggregate}/metadados"
    url = f"{baseurl}{query}"
    codes = get_json(url)["nivelTerritorial"]["Administrativo"]
    df = pd.DataFrame({"codes": codes})
    df["parameters"] = [locations_dict.get(code) for code in df.codes]
    return df


def list_periods(aggregate):
    """
    Function to list periods of a given aggregate.

    Parameters
    ----------
    aggregate : int or str
        Aggregate's code.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the frequency, initial and final dates of an aggregate.

    Examples
    --------
    >>> ibge.list_periods(1419)
               valores
    frequencia  mensal
    inicio      201201
    fim         201911
    """
    periods = get_metadata(aggregate).loc["periodicidade"][0]
    return pd.DataFrame(periods.values(), index=periods.keys(), columns=["valores"])


def list_classifications(aggregate, *search, **searches):
    """
    Function to list all classification of a given aggregate.

    Parameters
    ----------
    aggregate : int or str
        Aggregate's code.

    *search
        Strings to search in categories' names.

    **searches
        Strings to search in other columns.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the classifications and their categories.

    Examples
    --------
    >>> ibge.list_classifications(1419).head()
         id                                     nome unidade  nivel classificacao_id                      classificacao_nome
    0  7169                             Índice geral    None     -1              315  Geral, grupo, subgrupo, item e subitem
    1  7170                  1.Alimentação e bebidas    None     -1              315  Geral, grupo, subgrupo, item e subitem
    2  7171              11.Alimentação no domicílio    None     -1              315  Geral, grupo, subgrupo, item e subitem
    3  7172  1101.Cereais, leguminosas e oleaginosas    None     -1              315  Geral, grupo, subgrupo, item e subitem
    4  7173                            1101002.Arroz    None     -1              315  Geral, grupo, subgrupo, item e subitem
    """
    classifications = get_metadata(aggregate).loc["classificacoes"][0]
    df = pd.io.json.json_normalize(
        classifications, "categorias", meta=["id", "nome"], meta_prefix="classificacao_"
    )
    if search or searches:
        return search_df(df, search, searches).reset_index(drop=True)
    return df


def list_states(*search, **searches):
    """
    Function to list all states and their codes.

    Parameters
    ----------
    aggregate : int or str
        Aggregate's code.

    *search
        Strings to search in states' names.

    **searches
        Strings to search in other columns.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the states and their codes.

    Examples
    --------
    >>> ibge.list_states().head()
       id sigla      nome  regiao_id regiao_sigla regiao_nome
    0  11    RO  Rondônia          1            N       Norte
    1  12    AC      Acre          1            N       Norte
    2  13    AM  Amazonas          1            N       Norte
    3  14    RR   Roraima          1            N       Norte
    4  15    PA      Pará          1            N       Norte
    """
    return list_regions_helper("estados", search, searches)


def list_macroregions(*search, **searches):
    """
    Function to list all macroregions and their codes.

    Parameters
    ----------
    *search
        Strings to search in macroregions' names.

    **searches
        Strings to search in other columns.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the macroregions and their codes.

    Examples
    --------
    >>> ibge.list_macroregions()
       id sigla          nome
    0   1     N         Norte
    1   2    NE      Nordeste
    2   3    SE       Sudeste
    3   4     S           Sul
    4   5    CO  Centro-Oeste
    """
    return list_regions_helper("regioes", search, searches)


def list_cities(*search, **searches):
    """
    Function to list all cities and their codes.

    Parameters
    ----------
    *search
        Strings to search in cities' names.

    **searches
        Strings to search in other columns.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the cities and their codes.

    Examples
    --------
    >>> ibge.list_cities(UF_nome="Rio de Janeiro").head()
               id                nome  microrregiao_id       microrregiao_nome  mesorregiao_id     mesorregiao_nome  UF_id UF_sigla         UF_nome  regiao_id regiao_sigla regiao_nome
    3175  3300100      Angra dos Reis            33013     Baía da Ilha Grande            3305       Sul Fluminense     33       RJ  Rio de Janeiro          3           SE     Sudeste
    3176  3300159             Aperibé            33002  Santo Antônio de Pádua            3301  Noroeste Fluminense     33       RJ  Rio de Janeiro          3           SE     Sudeste
    3177  3300209            Araruama            33010                   Lagos            3304             Baixadas     33       RJ  Rio de Janeiro          3           SE     Sudeste
    3178  3300225               Areal            33005               Três Rios            3303    Centro Fluminense     33       RJ  Rio de Janeiro          3           SE     Sudeste
    3179  3300233  Armação dos Búzios            33010                   Lagos            3304             Baixadas     33       RJ  Rio de Janeiro          3           SE     Sudeste
    """
    return list_regions_helper("municipios", search, searches)


def list_microregions(*search, **searches):
    """
    Function to list all microregions.

    Parameters
    ----------
    *search
        Strings to search in microregions' names.

    **searches
        Strings to search in other columns.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with metadata about the microregions.

    Examples
    --------
    >>> ibge.list_microregions("Rio", mesorregiao_nome="Rio")
            id                   nome  mesorregiao_id                 mesorregiao_nome  UF_id UF_sigla         UF_nome  regiao_id regiao_sigla regiao_nome
    348  33018         Rio de Janeiro            3306  Metropolitana do Rio de Janeiro     33       RJ  Rio de Janeiro          3           SE     Sudeste
    352  35004  São José do Rio Preto            3501            São José do Rio Preto     35       SP       São Paulo          3           SE     Sudeste
    """
    return list_regions_helper("microrregioes", search, searches)


def list_mesoregions(*search, **searches):
    """
    Function to list all mesoregions and their codes.

    Parameters
    ----------
    *search
        Strings to search in mesoregions' names.

    **searches
        Strings to search in other columns.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with metadata about the mesoregions.

    Examples
    --------
    >>> ibge.list_mesoregions().head()
         id               nome  UF_id UF_sigla   UF_nome  regiao_id regiao_sigla regiao_nome
    0  1101    Madeira-Guaporé     11       RO  Rondônia          1            N       Norte
    1  1102  Leste Rondoniense     11       RO  Rondônia          1            N       Norte
    2  1201      Vale do Juruá     12       AC      Acre          1            N       Norte
    3  1202       Vale do Acre     12       AC      Acre          1            N       Norte
    4  1301   Norte Amazonense     13       AM  Amazonas          1            N       Norte
    """
    return list_regions_helper("mesorregioes", search, searches)


# vi: nowrap



from pandas import DataFrame
from .request import get_json


def bcb_metadata_to_df(url):
    """
    Auxiliary function to request metadata
    from BCB database and turn the returned
    JSON into a DataFrame.

    Parameters
    ----------
    url : str
        Url to be requested.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with metadatas in its columns.
    """
    json = get_json(url)["result"]["results"][0]
    return DataFrame.from_dict(json, orient="index", columns=["values"])


def ipea_metadata_to_df(url):
    """
    Auxiliary function to request metadata
    from IPEA database and turn the returned
    JSON into a DataFrame.

    Parameters
    ----------
    url : str
        Url to be requested.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with metadatas in its columns.
    """
    json = get_json(url)["value"][0]
    return DataFrame.from_dict(json, orient="index", columns=["values"])


def ibge_metadata_to_df(url):
    """
    Auxiliary function to request metadata
    from IBGE database and turn the returned
    JSON into a DataFrame.

    Parameters
    ----------
    url : str
        Url to be requested.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with metadatas in its columns.
    """
    json = get_json(url)
    return DataFrame.from_dict(json, orient="index", columns=["values"])



from pandas import concat, to_datetime

from .helpers.dates import parse_dates
from .helpers.utils import collect_codes_and_names
from .helpers.response import bcb_json_to_df
from .helpers.searching import bcb_get_search_results
from .helpers.metadata import bcb_metadata_to_df


def get_serie(code, name=None, start=None, end=None, last_n=None):
    """
    Auxiliary function to return a single time series
    from BCB database.
    """
    assert isinstance(code, (str, int)), "Not a valid code format."
    baseurl = f"https://api.bcb.gov.br/dados/serie/bcdata.sgs.{code}/dados"
    if last_n:
        url = f"{baseurl}/ultimos/{last_n}?formato=json"
    else:
        start, end = parse_dates(start, end, api="bcb")
        dates = f"&dataInicial={start}" if start else start
        dates += f"&dataFinal={end}" if end else end
        url = f"{baseurl}?format=json{dates}"
    return bcb_json_to_df(url, code, name)


def get_series(*codes, start=None, end=None, last_n=None, **kwargs):
    """
    Get multiple series into a DataFrame.

    Parameters
    ----------
    codes : dict, str, int
        Dictionary like {"name1": cod1, "name2": cod2}
        or a bunch of code numbers, e.g. cod1, cod2.

    start : str, optional
        Initial date, month or day first.

    end : str, optional
        End date, month or day first.

    last_n : int, optional
        Ignore other arguments and get last n observations.

    **kwargs
        Passed to pandas.concat.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with series' values.

    Examples
    --------
    >>> bcb.get_series({"Spread": 20786}, start="02-2018", end="072018")
                Spread
    Date
    2018-02-01   33.97
    2018-03-01   33.66
    2018-04-01   33.03
    2018-05-01   30.92
    2018-06-01   29.43
    2018-07-01   29.39
    """
    codes, names = collect_codes_and_names(*codes)
    df = concat(
        (get_serie(code, name, start, end, last_n) for code, name in zip(codes, names)),
        axis="columns",
        sort=True,
        **kwargs,
    )
    return df.sort_index()


def search(*search, rows=10, start=1):
    """
    Search for a name in the SGS database.

    Parameters
    ----------
    rows : int, default 10
        How many results to show.

    start : int, default 1
        From which row to start showing the results.

    *search
        Arbitrary number of strings to search.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the search results.

    Examples
    --------
    >>> bcb.search("Atividade", "econômica", rows=5, start=2)
      codigo_sgs                                              title periodicidade    unidade_medida
    0      27738  Saldo das operações de crédito por atividade e...        mensal  Milhões de reais
    1      27742  Saldo das operações de crédito por atividade e...        mensal  Milhões de reais
    2      22039  Saldo das operações de crédito por atividade e...        mensal  Milhões de reais
    3      22041  Saldo das operações de crédito por atividade e...        mensal  Milhões de reais
    4      22027  Saldo das operações de crédito por atividade e...        mensal  Milhões de reais
    """
    baseurl = "https://dadosabertos.bcb.gov.br/api/3/action/package_search?"
    params = f"q={search[0]}&rows={rows}&start={start}&sort=score desc"
    filter_params = ""
    if len(search) > 1:
        filter_params = f"&fq={'+'.join([name for name in search[1:]])}"
    url = f"{baseurl}{params}{filter_params}"
    return bcb_get_search_results(url)


def get_metadata(code):
    """
    Get metadata of a BCB's time series.

    Parameters
    ----------
    code : str
        Time series' code.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with metadata values.

    Examples
    --------
    >>> bcb.get_metadata(20786).head()
                                                                        values
    referencias              <P><A href="http://www.bcb.gov.br/estatisticas...
    license_title            Licença Aberta para Bases de Dados (ODbL) do O...
    maintainer                  Banco Central do Brasil/Departamento Econômico
    relationships_as_object                                                 []
    vcge                     Política Econômica [http://vocab.e.gov.br/2011...
    """
    baseurl = "https://dadosabertos.bcb.gov.br/api/3/action/package_search?"
    params = f"fq=codigo_sgs:{code}"
    url = f"{baseurl}{params}"
    return bcb_metadata_to_df(url)


# vi: nowrap



import re

from pandas import concat, to_datetime
from .helpers.utils import collect_codes_and_names
from . import bcb, ipea


def get_series(*codes, start=None, end=None, **kwargs):
    """
    Get multiple series from both BCB or IPEA.

    Parameters
    ----------
    codes : dict, str, int
        Dictionary like {"name1": cod1, "name2": cod2}
        or a bunch of code numbers, e.g. cod1, cod2.

    start : str, optional
        Initial date, month or day first.

    end : str, optional
        End date, month or day first.

    last_n : int, optional
        Ignore other arguments and get last n observations.

    **kwargs
        Passed to pandas.concat.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with series' values.

    Examples
    --------
    >>> seriesbr.get_series("BM12_CRLIN12", 20786, start="2015", end="2015")
                BM12_CRLIN12  20786
    Date
    2015-01-01          4.41  26.91
    2015-02-01          4.42  27.95
    2015-03-01          4.38  27.72
    2015-04-01          4.57  28.93
    2015-05-01          4.68  29.61
    2015-06-01          4.59  30.31
    2015-07-01          4.77  31.24
    2015-08-01          4.91  31.65
    2015-09-01          4.92  31.49
    2015-10-01          5.02  32.64
    2015-11-01          5.22  33.31
    2015-12-01          5.28  31.64
    """
    codes, names = collect_codes_and_names(*codes)
    series = []
    for code, name in zip(codes, names):
        if re.search(r"^\d+$", str(code)):
            df = bcb.get_serie(code, name, start, end)
        else:
            df = ipea.get_serie(code, name, start, end)
        series.append(df)
    return concat(series, axis="columns", **kwargs)



import re
import pandas as pd


def collect_codes_and_names(*args):
    """
    Auxiliary function to help label columns of a
    DataFrame.

    If it finds a dictionary, it will use its keys
    as labels and values as series' codes.

    Otherwise, the codes themselves will serve
    as columns' labels.

    Returns
    -------
    tuple
        First element with codes and second element
        with names.
    """
    codes, names = [], []
    for arg in args:
        if isinstance(arg, dict):
            for key, val in arg.items():
                codes.append(val)
                names.append(key)
        else:
            codes.append(arg)
            names.append(arg)
    return codes, names


def cat(something, sep):
    """
    Auxiliary function to join an iterable
    delimited by sep, forcing the join by
    coercing the items to be strings.

    Parameters
    ----------
    something : list
        List to be joined.

    sep : str
        Delimiter.

    Returns
    -------
    str
        String with values in something joined
        by sep.
    """
    return sep.join(map(str, something)) if is_iterable(something) else something


def is_iterable(something):
    """
    Auxiliary function to test if something is
    an iterable (unless it's a str).

    Returns
    -------
    bool
        True if an iterable other than str,
        False otherwise.
    """
    try:
        iter(something)
    except TypeError:
        return False
    return True and not isinstance(something, str)


def search_df(df, name, other=[]):
    """
    Auxiliary function to search for regex
    in a column of a DataFrame. This is a
    wrapper around the DataFrame query method.

    It builds a string to be passed to
    the query method of a DataFrame.

    Parameters
    ----------

    df : pandas.DataFrame

    search : list of strings

    other : dict

    Raises
    ------
    ValueError
        If the user tries to search in a
        non-existent column.
    """
    df_cols = df.columns.tolist()

    # check if additional searched columns exists
    for col in other:
        if col not in df_cols:
            raise ValueError(f"{col} is not an existing column.")

    # build regex string to search variable name
    regex = build_regex(name)
    name_search = f"nome.str.contains('{regex}')"

    # build regexes to search additional fields, if any
    if other:
        other_searches = []
        for field, search in other.items():
            regex = build_regex(search)
            other_searches.append(f"{field}.str.contains('{regex}')")
        other_searches = " and " + " and ".join(other_searches)
        return df.query(name_search + other_searches, engine="python")

    return df.query(name_search, engine="python")


def build_regex(strings):
    """
    Build regex by joining strings with '|' but
    only if it is an iterable other than a str.
    """
    # (?iu) sets unicode and ignore case flags
    flags = r'(?iu)'
    if is_iterable(strings):
        return f"{flags}{'|'.join(map(str, strings))}"
    return f"{flags}{strings}"


def clean_json(json):
    """
    Helper function to turn JSON into a DataFrame
    and clean its columns names.
    """
    df = pd.io.json.json_normalize(json, sep='_')
    df = df.rename(lambda x: '_'.join(re.split(r'_', x)[-2:]), axis='columns')
    return df

# vi: nowrap



from pandas import DataFrame
from .request import get_json
from .utils import clean_json, search_df


# IBGE


def list_regions_helper(region, search, searches):
    """
    Auxiliary function to list information
    about a given location in IBGE's database.

    Parameters
    ----------
    region : str
        Kind of region to list (macroregion, state etc.)

    search : list
        Strings to search in names column.

    searches : dict
        Strings to search in other columns.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the regions, filtered if specified.
    """
    url = f"https://servicodados.ibge.gov.br/api/v1/localidades/{region}"
    json = get_json(url)
    df = clean_json(json)
    if search or searches:
        return search_df(df, search, searches)
    return df


# IPEA


def list_metadata_helper(resource_path):
    """
    Auxiliary function to request metadata
    information about a IPEA's time series.

    Parameters
    ----------
    resource_path : str
        Which kind of information to be
        requested, e.g. themes, countries.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the specified resource.
    """
    baseurl = "http://www.ipeadata.gov.br/api/odata4/"
    url = f"{baseurl}{resource_path}"
    json = get_json(url)["value"]
    return DataFrame(json)



import setuptools
import distutils

long_description = """
SeriesBR - A Python package to interact with brazilian time series databases

It has functions to get time series data from Instituto de Pesquisa Econômica
Aplicada (IPEA), Banco Central do Brasil (BCB) and Instituto Brasileiro de
Geografia e Estatística (IBGE) databases into a DataFrame.

Learn more about it in our [repository](https://github.com/phelipetls/seriesbr)
or [documentation](https://seriesbr.readthedocs.io).
"""

setuptools.setup(
    name="seriesbr",
    version="0.1.6",
    author="Phelipe Teles",
    author_email="phelipe_teles@hotmail.com",
    description="A Python interface to interact with brazilian time series databases",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/phelipetls/seriesbr",
    packages=setuptools.find_packages(),
    install_requires=["pandas", "requests"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    project_urls={
        "Documentation": "https://seriesbr.readthedocs.io/",
    },
)



from pandas import concat, DataFrame, to_datetime

from .helpers.dates import parse_dates
from .helpers.utils import collect_codes_and_names
from .helpers.lists import list_metadata_helper
from .helpers.response import ipea_json_to_df
from .helpers.metadata import ipea_metadata_to_df
from .helpers.searching import ipea_get_search_results
from .helpers.ipea_metadata_list import ipea_metadata_list
from .helpers.url import (
    ipea_make_select_query,
    ipea_make_filter_query,
    ipea_make_date_query,
)


def get_serie(code, name=None, start=None, end=None):
    """
    Auxiliary function to return a single time series
    from IPEA database.
    """
    assert isinstance(code, str), "Not a valid code format."
    baseurl = "http://ipeadata2-homologa.ipea.gov.br/api/v1/"
    resource_path = f"ValoresSerie(SERCODIGO='{code}')"
    select = "?$select=VALDATA,VALVALOR"
    start, end = parse_dates(start, end, api="ipea")
    dates = ipea_make_date_query(start, end)
    url = f"{baseurl}{resource_path}{select}{dates}"
    return ipea_json_to_df(url, code, name)


def get_series(*codes, start=None, end=None, **kwargs):
    """
    Get multiple series into a DataFrame.

    Parameters
    ----------
    codes : dict, str
        Dictionary like {"name1": code1, "name2": code2}
        or a bunch of code strings, e.g. code1, code2.

    start : str
        Initial date, month or day first.

    end : str
        End date, month or day first.

    **kwargs
        Passed to pandas.concat.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with series' values.

    Examples
    --------
    >>> ipea.get_series("BM12_TJOVER12", "CAGED12_SALDO12", start="2018", end="03-2018")
                BM12_TJOVER12  CAGED12_SALDO12
    Date
    2018-01-01           0.58          77822.0
    2018-02-01           0.47          61188.0
    2018-03-01           0.53          56151.0
    """
    codes, names = collect_codes_and_names(*codes)
    df = concat(
        (get_serie(code, name, start, end) for code, name in zip(codes, names)),
        axis="columns",
        sort=True,
        **kwargs,
    )
    return df


def search(*SERNOME, **metadatas):
    """
    Function to search in IPEA's database.

    Parameters
    ----------
    *SERNOME
        String(s) to look up for in a series' name.

    **metadatas
        Keyword arguments where parameter is a valid metadata
        and value a str or list of str.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the search results.

    Examples
    --------
    >>> ipea.search("Taxa", "Juros", "Selic", "recursos livres", PERNOME="mensal", UNINOME="%").head()
            SERCODIGO                                            SERNOME PERNOME   UNINOME
    0    BM12_CRDTJ12  Operações de crédito - recursos direcionados -...  Mensal  (% a.a.)
    1  BM12_CRDTJPF12  Operações de crédito - recursos direcionados -...  Mensal  (% a.a.)
    2  BM12_CRDTJPJ12  Operações de crédito - recursos direcionados -...  Mensal  (% a.a.)
    3    BM12_CRLIN12  Operações de crédito - recursos livres - inadi...  Mensal       (%)
    4  BM12_CRLINPF12  Operações de crédito - recursos livres - inadi...  Mensal       (%)
    """
    baseurl = "http://ipeadata2-homologa.ipea.gov.br/api/v1/"
    resource_path = "Metadados"
    select_query = ipea_make_select_query(metadatas)
    filter_query = ipea_make_filter_query(SERNOME, metadatas)
    url = f"{baseurl}{resource_path}{select_query}{filter_query}"
    return ipea_get_search_results(url)


def get_metadata(code):
    """
    Get metadata of a series specified
    by the a code.

    Parameters
    ----------
    code : int or str

    Returns
    -------
    pandas.DataFrame
        A DataFrame with series' metadata.

    Examples
    --------
    >>> ipea.get_metadata("BM12_TJOVER12").head()
                                                               values
    SERCODIGO                                           BM12_TJOVER12
    SERNOME                              Taxa de juros - Over / Selic
    SERCOMENTARIO   Quadro: Taxas de juros efetivas.  Para 1974-19...
    SERATUALIZACAO                      2019-12-17T05:06:00.793-02:00
    BASNOME                                            Macroeconômico
    """
    baseurl = "http://ipeadata2-homologa.ipea.gov.br/api/v1/"
    resource_path = f"Metadados('{code}')"
    url = f"{baseurl}{resource_path}"
    return ipea_metadata_to_df(url)


def list_themes():
    """
    Function to list all themes available
    in the database.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with all available themes
        in IPEA's database.

    Examples
    --------
    >>> ipea.list_themes().head()
        TEMCODIGO  TEMCODIGO_PAI                  TEMNOME
    0          28            NaN             Agropecuária
    1          23            NaN       Assistência social
    2          10            NaN    Balanço de pagamentos
    3           7            NaN                   Câmbio
    4           5            NaN        Comércio exterior
    """
    return list_metadata_helper("Temas")


def list_countries():
    """
    Function to list all countries available
    in the database.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with all available countries
        in IPEA's database.

    Examples
    --------
    >>> ipea.list_countries().head()
      PAICODIGO         PAINOME
    0       ZAF   África do Sul
    1       DEU        Alemanha
    2      LATI  América Latina
    3       AGO          Angola
    4       SAU  Arábia Saudita
    """
    return list_metadata_helper("Paises")


def list_metadata():
    """
    Function to list all valid metadatas and their description.

    Returns
    -------
    pandas.DataFrame

    Examples
    --------
    >>> ipea.list_metadata().head()
                       Description
    SERNOME                   Name
    SERCODIGO                 Code
    PERNOME              Frequency
    UNINOME    Unit of measurement
    BASNOME           Basis's name
    """
    return DataFrame.from_dict(
        ipea_metadata_list, orient="index", columns=["Description"]
    )


# vi: nowrap
